def cluster_genes(adata, haystack_result, method="kmeans", n_clusters=None, n_genes=100, random_state=None):
  """
  Cluster singleCellHaystack results into gene modules.

  :param adata: AnnData object.
  :param haystack_result: singleCellHaystack result.
  :param method: clustering method. Currently only sklearn.cluster.KMeans implemented.
  :param n_clusters: number of clusters for KMeans.
  :param n_genes: number of top genes to cluster.
  :param random_state: number to set random_state in clustering method.

  """

  import numpy as np
  from ._haystack import calculate_P_dist
  from sklearn.cluster import KMeans
  from pandas import DataFrame
  from scipy.sparse import isspmatrix
  #import scipy.stats as ss

  #grid_points = r["info"]["grid.points"]
  #grid_dist = calculate_dist_to_cells(coord, grid_points)
  #density = calculate_density(grid_dist)
  density = haystack_result["info"]["grid_density"]
  ngrid_points = density.shape[1]

  sum = haystack_result["results"]
  sum = sum.head(n_genes)
  genes = sum["gene"]
  adata = adata[:, genes]
  expression = adata.X
  ngenes = expression.shape[1]

  if (isspmatrix(expression)):
    expression = expression.tocsc()

  # FIXME: vectorize this computation.
  scores = np.zeros([ngenes, ngrid_points])
  for k in range(ngenes):
    scores[k, :] = calculate_P_dist(density, expression[:, k])

  if method == "kmeans":
    res = KMeans(n_clusters=n_clusters, random_state=random_state, n_init=10).fit(scores)
    clusters = res.labels_

  if method == "hclust":
    cluters = None

  res = DataFrame({
    "gene": adata.var_names,
    "cluster": clusters
  })
  
  return res

def plot_gene_clusters(adata, gene_clusters, basis=None, ncols=4, figsize=None, color_map="coolwarm"):
  import numpy as np
  import matplotlib.pyplot as plt

  if basis is None:
    basis_key = adata.obsm_keys()[0]
  else:
    if basis in adata.obsm_keys():
      basis_key = basis
    elif f"X_{basis}" in adata.obsm_keys():
      basis_key = f"X_{basis}"

  if basis_key is None:
    print("No coordinates found!")

  coord = adata.obsm[basis_key]

  clusters = gene_clusters.cluster.unique()
  scores = np.zeros((adata.n_obs, len(clusters)))
  for k in range(len(clusters)):
    genes = gene_clusters[gene_clusters.cluster == clusters[k]].gene.tolist()
    m = adata[:, genes].X
    scores[:,k] = np.squeeze(np.mean(m, axis=1))

  import matplotlib.pyplot as plt
  nclusters = scores.shape[1]
  nrows = int(np.ceil(nclusters/ncols))

  fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)

  for i in range(nrows):
    for j in range(ncols):
      ax[i, j].grid(False)
      ax[i, j].set_axis_off()

  crow=0
  ccol=0
  for k in range(nclusters):
    im = ax[crow, ccol].scatter(coord[:,0], coord[:,1], s=4, c=scores[:,k], cmap=color_map)
    ax[crow, ccol].set_title("Module: " + str(k))
    ax[crow, ccol].set_axis_on()
    ax[crow, ccol].set_xticks([])
    ax[crow, ccol].set_yticks([])
    ax[crow, ccol].set_xlabel("UMAP1")
    ax[crow, ccol].set_ylabel("UMAP2")
    fig.colorbar(im, ax=ax[crow,ccol])
  #  print("crow: " + str(crow) + ", ccol: " + str(ccol))
    ccol = ccol + 1
    if (ccol > ncols - 1):
      crow+=1
      ccol=0

  fig.tight_layout()

  return scores



import numpy as np

def estimate_spline_param(x, y, method="bs"):
  from sklearn.preprocessing import SplineTransformer
  from sklearn.preprocessing import FunctionTransformer
  from sklearn.linear_model import LinearRegression
  from sklearn.pipeline import Pipeline
  from sklearn.model_selection import GridSearchCV
  from patsy import cr

  if method == "bs":
    pip = Pipeline([
      ["transformer", SplineTransformer()],
      ["estimator", LinearRegression()]
    ])

    min_knots = 2
    max_knots = 10
    min_degree = 1
    max_degree = 5
    param = {
      "transformer__n_knots": list(range(min_knots, max_knots)),
      "transformer__degree": list(range(min_degree, max_degree))
    }

    cv = GridSearchCV(pip, param, cv=10)
    cv_res = cv.fit(x, y)

    info = {
      "method": "bs",
      "n_knots": cv_res.best_params_["transformer__n_knots"],
      "degree": cv_res.best_params_["transformer__degree"]
    }

  if method == "ns":
    NaturalSplineTransformer = FunctionTransformer(cr)
    pip = Pipeline([
      ["transformer", NaturalSplineTransformer],
      ["estimator", LinearRegression()]
    ])

    # Set dict of degrees of freedom for CV.
    min_df = 3
    max_df = 10
    df = []
    for k in range(min_df, max_df):
      df.append({"df": k})

    param = {
      "transformer__kw_args": df
    }

    cv = GridSearchCV(pip, param, cv=10)
    cv_res = cv.fit(x, y)

    info = {
      "method": "ns",
      "df": cv_res.best_params_["transformer__kw_args"]["df"]
    }

  return info

def calculate_KLD_fit(x, y, method="bs"):
  from sklearn.preprocessing import SplineTransformer
  from sklearn.preprocessing import FunctionTransformer
  from sklearn.linear_model import LinearRegression
  from sklearn.pipeline import Pipeline
  from patsy import cr

  info = estimate_spline_param(x, y, method=method)
  if method == "bs":
    n_knots = info["n_knots"]
    degree = info["degree"]

    pip = Pipeline([
      ["transformer", SplineTransformer(n_knots=n_knots, degree=degree)],
      ["estimator", LinearRegression()]
    ])
    model = pip.fit(x, y)
    y_hat = model.predict(x)

  if method == "ns":
    df = info["df"]

    NaturalSplineTransformer = FunctionTransformer(cr, kw_args={"df": df})
    pip = Pipeline([
      ["transformer", NaturalSplineTransformer],
      ["estimator", LinearRegression()]
    ])
    model = pip.fit(x, y)
    y_hat = model.predict(x)

  return {
    "model": model,
    "spline": info,
    "y_hat": y_hat
  }

def calculate_Pval(KLD, KLD_rand, cv, cv_rand, method="bs", verbose=False):
  from sklearn.preprocessing import SplineTransformer
  from sklearn.linear_model import LinearRegression
  from sklearn.pipeline import Pipeline
  from patsy import cr
  from scipy.stats import norm

  if (verbose):
    print("> calculating P values ...")

  KLD_log = np.log(KLD)
  KLD_rand_log = np.log(KLD_rand)
  KLD_rand_mean = np.mean(KLD_rand_log, axis=1)
  KLD_rand_sd = np.std(KLD_rand_log, axis=1)
  cv_log = np.log(cv).reshape(-1, 1)
  cv_rand_log = np.log(cv_rand).reshape(-1, 1)

  KLD_rand_mean_fit = calculate_KLD_fit(cv_rand_log, KLD_rand_mean, method=method)
  KLD_rand_sd_fit = calculate_KLD_fit(cv_rand_log, KLD_rand_sd, method=method)

  KLD_rand_mean_model = KLD_rand_mean_fit["model"]
  KLD_rand_sd_model = KLD_rand_sd_fit["model"]

  KLD_mean = KLD_rand_mean_model.predict(cv_log)
  KLD_sd = KLD_rand_sd_model.predict(cv_log)

  logpval = norm.logsf(KLD_log, loc=KLD_mean, scale=KLD_sd)/np.log(10)
  pval = 10 ** logpval

  return {
    "pval": pval,
    "logpval": logpval,
    "method": method,
    "CV": cv_rand_log,
    "rand_mean": KLD_rand_mean,
    "rand_sd": KLD_rand_sd,
    "rand_mean_model": KLD_rand_mean_model,
    "rand_sd_model": KLD_rand_sd_model,
    "rand_mean_spline": KLD_rand_mean_fit["spline"],
    "rand_sd_spline": KLD_rand_sd_fit["spline"],
    "rand_mean_hat": KLD_rand_mean_fit["y_hat"],
    "rand_sd_hat": KLD_rand_sd_fit["y_hat"]
  }



import numpy as np
from scipy.sparse import isspmatrix
from tqdm import tqdm

# density is density as calculated by calculate_density.
def calculate_Q_dist(density, pseudo=1e-300, verbose=False):
  if (verbose):
    print("> calculating Q dist ...")

  Q = np.sum(density, axis=0)
  Q = Q + pseudo
  Q = Q / np.sum(Q)
  return Q

# density (array): ncells x ngrid_points.
# expression (vector): ncells.
# P (vector): ngrid_points.
def calculate_P_dist(density, weights, pseudo=1e-300):
  from anndata import AnnData
  from numpy import ndarray

  if (isspmatrix(weights)):
    index = weights.nonzero()[0]
    P = density[index, :] * weights.data.reshape(-1,1)

  if (isinstance(weights, ndarray)):
    P = density * weights.reshape(-1,1)

  P = np.sum(P, 0)
  P = P + pseudo
  P = P / np.sum(P)
  return P

def calculate_KLD(density, weights, Q, pseudo=1e-300, verbose=False):

  if (isspmatrix(weights)):
    weights = weights.tocsc()

  ngenes = weights.shape[1]

  if (verbose):
    print("> calculating KLD for " + str(ngenes) + " features ...")
    pbar = tqdm(total=ngenes)

  # FIXME: vectorize this computation.
  res = np.zeros(ngenes)
  for k in range(ngenes):
    P = calculate_P_dist(density, weights[:, k])
    res[k] = np.sum(P * np.log(P / Q))
    if (verbose):
      pbar.update(n=1)

  return res



import numpy as np

# x is a matrix of PCA or other embeddings coordinates.
def calculate_grid_points(x, ngrid_points, random_state=None, verbose=False):
  from sklearn.cluster import KMeans

  if (verbose):
    print("> calculating grid points ...")

  res = KMeans(n_clusters=ngrid_points, random_state=random_state, n_init=10).fit(x)
  return res.cluster_centers_

def calculate_dist_to_cells(coord, grid_points, verbose=False):
  from scipy.spatial import distance_matrix

  if (verbose):
    print("> calculating distance to cells ...")

  return distance_matrix(coord, grid_points)

# dist is a distance matrix, calculated form grid points and embedding
# coordinates.
def calculate_density(dist, verbose=False):
  if (verbose):
    print("> calculating densities ...")

  bandwidth = np.median(np.min(dist, 1))
  dist_norm = dist / bandwidth
  return np.exp(-dist_norm * dist_norm / 2)



import numpy as np
from scipy.sparse import isspmatrix
from random import sample
from tqdm import tqdm

def select_genes_to_randomize(x, ngenes=100, method="heavytails", tail=10, verbose=False):
  if (verbose):
    print("> selecting genes to randomize ...")

  index = np.argsort(x)

  if (method == "uniform"):
    return index[np.linspace(0, index.shape[0]-1, ngenes, dtype="int")]

  if (method == "heavytails"):
    ls = index[:tail]
    rs = index[-tail:]
    ngenes = ngenes - tail * 2
    ms = index[np.linspace(tail, index.shape[0]-tail-1, ngenes, dtype="int")]
    return np.concatenate([ls, ms, rs])

def randomize_KLD(density, expression, Q, n_randomizations=100, pseudo=1e-300, verbose=False):
  from ._haystack import calculate_KLD

  if (verbose):
    print("> calculating randomized KLD ...")
    pbar = tqdm(total=n_randomizations)

  if (isspmatrix(expression)):
    expression = expression.tocsc()

  ncells=expression.shape[0]
  ngenes=expression.shape[1]

  KLD_rand=np.zeros([ngenes, n_randomizations])

  for n in range(n_randomizations):
    shuffled_cells = sample(range(ncells), ncells)
    KLD_rand[:, n] = calculate_KLD(density, expression[shuffled_cells, :], Q)
    if (verbose):
      pbar.update(n=1)

  return KLD_rand



from ._randomization import *
from ._pvalue import *
from ._grid import *
from ._kld import *

from scipy.sparse import isspmatrix

# haystack
# Main function. Can accept AnnData, numpy array and scipy sparse matrices.
# Does not accept numpay matrix objects.
def haystack(x, coord, features=None, scale_coords=True, ngrid_points=100,
    n_genes_to_randomize=100, select_genes_randomize_method="heavytails",
    spline_method="bs", n_randomizations=100, grid_points=None, pseudo=1e-300, verbose=True):

  """
  Runs singleCellHaystack.

  :param x: AnnData, numpy array or scipy sparse matrix.
  :param coord: a numpy array with 1D pseudotime, 2D or 3D spatial coordinates or an embedding with any number of dimansions.
  :return: A list with singleCellHaystack results.
  :rtype: list

  """
  from anndata import AnnData
  from numpy import ndarray

  if isinstance(x, AnnData) and isinstance(coord, str):
    res = haystack_adata(adata=x, basis=coord, dims=None, scale_coords=scale_coords,
        ngrid_points=ngrid_points, n_genes_to_randomize=n_genes_to_randomize,
        select_genes_randomize_method=select_genes_randomize_method, spline_method=spline_method,
        n_randomizations=n_randomizations, grid_points=grid_points, pseudo=pseudo, verbose=verbose)

  if (isinstance(x, ndarray) or isspmatrix(x)) and isinstance(coord, ndarray):
    res = haystack_array(weights=x, coord=coord, features=features, scale_coords=scale_coords,
        ngrid_points=ngrid_points, n_genes_to_randomize=n_genes_to_randomize,
        select_genes_randomize_method=select_genes_randomize_method, spline_method=spline_method,
        n_randomizations=n_randomizations, grid_points=grid_points, pseudo=pseudo, verbose=verbose)

  return(res)

# haystack_array
# Method for numpy array and scipy sparse matrix objects.
def haystack_array(weights, coord, features=None, scale_coords=True, ngrid_points=100,
    n_genes_to_randomize=100, select_genes_randomize_method="heavytails",
    spline_method="bs", n_randomizations=100, grid_points=None, pseudo=1e-300, verbose=True):

  from pandas import DataFrame
  from sklearn.preprocessing import StandardScaler

  if (verbose):
    print("> entering array method ...")

  exprs = weights

  # Features included?
  if features is None:
    features = np.array(range(exprs.shape[1]))

  # Scale coords.
  if scale_coords:
    if (verbose):
      print("> scaling coordinates ...")

    coord_mean = np.mean(coord, axis=0)
    coord_std = np.std(coord, axis=0)
    coord = (coord - coord_mean) / coord_std

  # Check for negative values.
  if (np.sum(exprs < 0).astype(bool)):
    print("_ERROR_ negative values in your data.")
    return(None)

  # filter genes with zero stdev.
  if (verbose):
    print("> calculating feature stds ...")
  scaler = StandardScaler(with_mean=False)
  scaler_fit = scaler.fit(exprs)
  exprs_sd = np.sqrt(scaler_fit.var_)
  exprs_mean = scaler_fit.mean_

  #exprs_sd = np.std(exprs, axis=0)
  sel_zero = exprs_sd == 0
  n_zero = np.sum(sel_zero)
  if (n_zero > 0):
    if (verbose):
      print("> removing", str(n_zero), "genes with zero variance ...")
    sel_nozero = np.invert(sel_zero)
  #   genes = genes[sel_nozero]
    exprs_sd = exprs_sd[sel_nozero] + pseudo
    exprs_mean = exprs_mean[sel_nozero] + pseudo
    exprs = exprs[:, sel_nozero]
    features = features[sel_nozero]

  ncells = exprs.shape[0]
  ngenes = exprs.shape[1]

  # Calculate KLD.
  if grid_points is None:
    grid_points = calculate_grid_points(coord, ngrid_points, verbose=verbose)
  grid_dist = calculate_dist_to_cells(coord, grid_points, verbose=verbose)
  grid_density = calculate_density(grid_dist, verbose=verbose)

  Q = calculate_Q_dist(grid_density, pseudo=pseudo, verbose=verbose)

  KLD = calculate_KLD(grid_density, exprs, Q, verbose=verbose)

  # Calculate CV
  if (verbose):
    print("> calculating feature's CV ...")
  exprs_cv = exprs_sd / exprs_mean

  genes_to_randomize = select_genes_to_randomize(exprs_cv, n_genes_to_randomize, method=select_genes_randomize_method, verbose=verbose)

  # Randomizations.
  KLD_rand = randomize_KLD(grid_density, exprs[:, genes_to_randomize], Q, n_randomizations=n_randomizations, verbose=verbose)

  # Calculate p.values:
  pvalData = calculate_Pval(KLD, KLD_rand, exprs_cv, exprs_cv[genes_to_randomize], method=spline_method, verbose=verbose)
  pval = pvalData["pval"]
  logpval = pvalData["logpval"]
  #pval_adj = pval * pval.size # Bonferroni correction
  logpval_adj = logpval + np.log10(logpval.size)
  logpval_adj = np.fmin(0, logpval_adj)
  pval_adj = 10 ** logpval_adj

  if (verbose):
    print("> done.")

  # Return results.
  df = DataFrame({
    "gene": features,
    "KLD": KLD,
    "pval": pval,
    "pval_adj": pval_adj,
    "logpval": logpval,
    "logpval_adj": logpval_adj
  })

  df = df.sort_values("logpval")

  info = {
    "grid_points": grid_points,
    "grid_dist": grid_dist,
    "grid_density": grid_density,
    "Q": Q,
    "KLD_rand": KLD_rand,
    "pval_info": pvalData,
    "genes_to_randomize": genes_to_randomize,
    "exprs_cv": exprs_cv,
  }

  return {
    "results": df,
    "info": info
  }

# haystack_adata
# method for AnnData objects.
def haystack_adata(adata, basis="pca", dims=None, scale_coords=True, ngrid_points=100,
    n_genes_to_randomize=100, select_genes_randomize_method="heavytails", spline_method="bs",
    n_randomizations=100, grid_points=None, pseudo=1e-300, verbose=True):

  if (verbose):
    print("> starting haystack ...")

  if basis in adata.obsm.keys():
    basis_key = basis
  elif f"X_{basis}" in adata.obsm.keys():
    basis_key = f"X_{basis}"

  coord = adata.obsm[basis_key]
  exprs = adata.X
  genes = adata.var_names

  # Check for negative values.
  if (np.sum(exprs < 0).astype(bool)):
    print("_ERROR_ negative values in your data. Hint: pass adata.raw.to_adata() or adata.layers[\"count\"]")
    return(None)

  if dims is not None:
    coord = coord[:, dims]

  res = haystack_array(exprs, coord, features=genes, scale_coords=scale_coords, ngrid_points=ngrid_points,
      n_genes_to_randomize=n_genes_to_randomize, select_genes_randomize_method=select_genes_randomize_method,
      spline_method=spline_method, n_randomizations=n_randomizations, grid_points=grid_points, pseudo=pseudo, verbose=verbose)
  return(res)



import matplotlib.pyplot as plt
import numpy as np

def plot_rand_fit(x, type="mean"):
  """
  Plot singleCellHaystack randomization information.

  :param x: singleCellHaystack result.
  :param type: one of "mean" or "sd".

  """

  data = x["info"]["pval_info"]

  x = data["CV"]
  method = data["method"]

  if type == "mean":
    y = data["rand_mean"]
    y_hat = data["rand_mean_hat"]

    if method == "bs":
      title = "fit: " + method + " n_knots: " + str(data["rand_mean_spline"]["n_knots"]) + " degree: " + str(data["rand_mean_spline"]["degree"])

    if method == "ns":
      title = "fit: " + method + " df: " + str(data["rand_mean_spline"]["df"])

  if type == "sd":
    y = data["rand_sd"]
    y_hat = data["rand_sd_hat"]

    if method == "bs":
      title = "fit: " + method + " n_knots: " + str(data["rand_sd_spline"]["n_knots"]) + " degree: " + str(data["rand_sd_spline"]["degree"])

    if method == "ns":
      title = "fit: " + method + " df: " + str(data["rand_sd_spline"]["df"])

  plt.scatter(x, y, color="black")
  plt.plot(x, y_hat, color="red")
  plt.xlabel("log CV")
  plt.ylabel(type)
  plt.title(title)
  plt.show()

def plot_pval_rank(x):
  """
  Plot singleCellHaystack p.value rank.

  :param x: singleCellHaystack result.

  """

  data = x["info"]["pval_info"]

  x = np.sort(data["logpval"])

  plt.plot(x, color="black")
  plt.xlabel("Rank")
  plt.ylabel("logpval")
  plt.show()

def plot_pval_hist(x):
  """
  Plot singleCellHaystack p.value histogram.

  :param x: singleCellHaystack result.

  """

  data = x["info"]["pval_info"]

  x = data["pval"]

  h = plt.hist(x, color="black", bins=30)
  plt.xlim([0, 1])
  plt.show()

def plot_compare_ranks(res1, res2, sort_by="logpval", xlabel=None, ylabel=None):
  """
  Plot the rank of two singleCellHaystack results.

  This plot facilitates comparing two singleCellHaystack results.

  :param res1: singleCellHaystack result.
  :param res2: singleCellHaystack resuot.
  :param sort_by: what column to use for sorting (rank).
  :param xlabel: label for x-axis (res1).
  :param ylabel: label for y-axis (res2).

  """

  import pandas as pd

  sum1 = res1["results"].sort_values(sort_by)
  sum2 = res2["results"].sort_values(sort_by)

  r1 = pd.DataFrame({
    "gene1": sum1.index,
    "rank1": list(range(sum1.index.shape[0])),
  })

  r2 = pd.DataFrame({
    "gene2": sum2.index,
    "rank2": list(range(sum2.index.shape[0])),
  })

  r1 = r1.sort_values("gene1")
  r1 = r1.reset_index(drop=True)

  r2 = r2.sort_values("gene2")
  r2 = r2.reset_index(drop=True)

  d = r1.join(r2)

  plt.plot(d.rank1, d.rank2, "bo", markersize=1)
  plt.xlabel(xlabel)
  plt.ylabel(ylabel)



""" 

This is a python implementation of `singleCellHaystack <https://github.com/alexisvdb/singleCellHaystack>`_.

"""

from anndata import AnnData

from ._haystack import *
from ._plot import *
from ._cluster import *
from ._data import *
